
; -----------------------------------------------------------------------------
; Real Mirrors! -- james @ hi-toro.com
; -----------------------------------------------------------------------------

; -----------------------------------------------------------------------------
; A list of ^2 texture sizes for 'black bar' workaround...
; -----------------------------------------------------------------------------

; This is necessary because cameras can only render to the back buffer. If you
; create a texture wider or higher than the screen mode, it'll be 'cut off'
; when rendered during UpdateMirror (this was the source of the black bars in
; certain situations before).

; The solution I've used is simply to create a texture that's smaller than the
; screen if you request one larger -- this means you can request large textures
; and they'll automatically be made smaller if necessary, avoiding the ugly
; black bars problem... so, let's create a bunch of 'power-of-2' texture sizes
; (1, 2, 4, 8... etc) to step through when needed:

Type TextureSize
	Field size
End Type

size = 1
For a = 1 To 20 ; 20 is overkill, but unarguably future-proof ;)
	ts.TextureSize = New TextureSize
	ts\size = size
	size = size * 2
Next

; -----------------------------------------------------------------------------
; Mirror type...
; -----------------------------------------------------------------------------

Type Mirror
	Field size#			; Mesh size for mirror/backing
	Field mirror		; Mirror mesh -- access using MirrorMesh (m.Mirror)
	Field backing		; Backing mesh -- access using BackingMesh (m.Mirror)
	Field texture		; Mirror texture handle
	Field texwidth		; Actual texture width after creation (may not be as requested)
	Field texheight		; Actual texture height after creation (may not be as requested)
	Field cam			; Mirror's camera (which grabs its image)
	Field camrange#		; Mirror's camera range
	Field dist#			; Mirror's range, beyond which it stops working
End Type

; -----------------------------------------------------------------------------
; Returns handle of mirror's mesh (for positioning, etc)...
; -----------------------------------------------------------------------------

; PARAMETERS [defaults in square brackets]:

; m.Mirror				-- previously created mirror

Function MirrorMesh (m.Mirror)
	Return m\mirror
End Function

; -----------------------------------------------------------------------------
; Returns handle of mirror's backing (might want to re-colour or texture it)...
; -----------------------------------------------------------------------------

; PARAMETERS [defaults in square brackets]:

; m.Mirror				-- previously created mirror

Function MirrorBacking (m.Mirror)
	Return m\backing
End Function

; -----------------------------------------------------------------------------
; Creates a mirror object (returns .Mirror type handle)...
; -----------------------------------------------------------------------------

; PARAMETERS [defaults in square brackets]:

; size#			[1]		-- size of mirror mesh
; texturesize	[256]	-- requested size of mirror's texture (128 or 256 recommended)
; flags			[0]		-- mirror's texture flags (0 [any gfx card] or 256 [high-end cards] recommended)
; clsr			[0]		-- ClsColor for mirror's camera (red value)
; clsg			[0]		-- ClsColor for mirror's camera (red value)
; clsb			[0]		-- ClsColor for mirror's camera (red value)
; dist#			[20]	-- mirror's range of effect
; camrange#		[1000]	-- range of mirror's camera (usually small, but planes, etc may be cut off)
; usingmask		[0]		-- set to True if you intend to apply a shape mask (see ApplyMirrorMask)

Function CreateRealMirror.Mirror (size# = 1, texturesize = 256, flags = 0, clsr = 0, clsg = 0, clsb = 0, dist# = 20, camrange# = 1000, usingmask = False)

	m.Mirror = New Mirror

	; If requested texture size is wider or higher than screen, get next size down...
	
	If (texturesize > GraphicsWidth ()) Or (texturesize > GraphicsHeight ())
		; Stepping backwards through TextureSize list...
		ts.TextureSize = Last TextureSize
		Repeat
			If (ts\size < GraphicsWidth ()) And (ts\size < GraphicsHeight ())
				texturesize = ts\size: Exit
			EndIf
			ts = Before ts
		Until ts = Null
	EndIf			
			
	m\size = size													; Mirror's mesh size
	m\texture = CreateTexture (texturesize, texturesize, flags)		; Mirror's texture
	ScaleTexture m\texture, -1, 1									; Flipped horizontally

	m\texwidth = TextureWidth (m\texture)
	m\texheight = TextureHeight (m\texture)

	m\mirror = MeshRect (0, m\size)									; Mirror mesh
	m\dist = dist													; Mirror's 'operating range'
	EntityAutoFade m\mirror, m\dist / 2.0, m\dist					; Fade out

	If usingmask Then TextureBlend m\texture, 1						; Needed if applying a shape mask
	EntityTexture m\mirror, m\texture, 0, 1							; Apply the mirror's texture to it

	m\backing = MeshRect (0, m\size)								; Mirror's backing mesh
	MoveEntity m\backing, 0, 0, 0.03								; Move it back a little
	EntityParent m\backing, m\mirror								; NOTE: Mirror mesh is parent!

	m\cam = CreateCamera (m\mirror)									; Mirror's camera (grabs view)
	m\camrange = camrange
	CameraRange m\cam, 0.1, m\camrange								; Camera's range (generally small, but infinite planes will look cut off with small values. The smaller the range, the less the camera has to render though.)
	CameraClsColor m\cam, clsr, clsg, clsb							; Camera's ClsColor (make same as your main camera)
	CameraViewport m\cam, 0, 0, m\texwidth, m\texheight				; Camera renders to size of mirror's texture
	CameraProjMode m\cam, 0											; No rendering until needed...

	EntityShininess m\mirror, 1										; Mirrors are shiny :)
	UpdateNormals m\mirror											; Dunno if this makes any difference here actually...

	Return m														; Return handle of type .Mirror

End Function

; -----------------------------------------------------------------------------
; Update specified mirror...
; -----------------------------------------------------------------------------

; PARAMETERS [defaults in square brackets]:

; usecam				-- the camera you are using in your game
; m.Mirror				-- previously created mirror

; NOTE: Assumes the current buffer is the BACK BUFFER! Set it manually before
; calling if you're on another buffer for some reason... but this will be
; fine in most cases.

Function UpdateMirror (usecam, m.Mirror)

	; Check the mirror exists...
	
	If m <> Null

		; Get distance from main camera to mirror...
		mirrordist# = EntityDistance (usecam, m\mirror)
		
		; Within mirror's range?
		
		If mirrordist < m\dist
		
			; Translate camera's world co-ordinates to xyz offset from mirror...			
			TFormPoint EntityX (usecam, 1), EntityY (usecam, 1), EntityZ (usecam, 1), 0, m\mirror
			
			; Get co-ordinates for mirror's camera (the one used to grab its image)...
			usecamx# = TFormedX ()
			usecamy# = TFormedY ()
			usecamz# = -TFormedZ ()			; [NOTE: NEGATIVE of main camera z!]

			; If the mirror's camera is now behind the mirror...
			
			If usecamz > 0

				; And in view of the main camera...
				
				If EntityInView (m\mirror, usecam)
				
					; Position the mirror's camera behind the mirror and point it at mirror...
					PositionEntity m\cam, usecamx, usecamy, usecamz
					PointEntity m\cam, m\mirror
					
					; Correct roll problem when pointing down...
					RotateEntity m\cam, EntityPitch (m\cam), EntityYaw (m\cam), 0
					
					; Adjust zoom of mirror's camera (this is one of the most important/hard parts to get!)...
					CameraZoom m\cam, mirrordist * (2.0 / m\size)
					CameraRange m\cam, mirrordist, m\camrange
					CameraProjMode usecam, 0			; Stop main camera rendering...
					CameraProjMode m\cam, 1				; Start mirror's camera rendering...
					RenderWorld							; Render what it sees to the current buffer...
					
					; Copy the rendered scene to the mirror's texture...
					CopyRect 0, 0, m\texwidth, m\texheight, 0, 0, 0, TextureBuffer (m\texture)
	
					CameraProjMode m\cam, 0				; Stop mirror's camera rendering...
					CameraProjMode usecam, 1			; Start main camera rendering again...
					
				EndIf
				
			EndIf
			
		EndIf
	
	EndIf
	
End Function

; -----------------------------------------------------------------------------
; Load and apply a mask image (from an image file) to the mirror...
; -----------------------------------------------------------------------------

; PARAMETERS [defaults in square brackets]:

; m.Mirror				-- previously created mirror
; maskfile$				-- alpha mask image file (eg. a BMP with a white circle on a black background)

Function ApplyMirrorMask (m.Mirror, maskfile$)
	mask = LoadTexture (maskfile$, 2)				; Load alpha mask
	If mask
		EntityTexture m\mirror, mask, 0, 0			; Apply to mirror texture, index 0
		EntityTexture m\backing, mask, 0, 0			; Apply to backing texture, index 0
	EndIf
End Function

; -----------------------------------------------------------------------------
; Create a square (just called internally by CreateRealMirror)...
; -----------------------------------------------------------------------------

Function MeshRect (parent = 0, size# = 1)
	newMesh = CreateMesh (parent)
	surface = CreateSurface (newMesh)
	size# = size / 2
	topLeft		= AddVertex (surface, -size, size, 0, 0, 0)
	topRight	= AddVertex (surface, size, size, 0, 1, 0)
	bottomRight	= AddVertex (surface, size, -size, 0, 1, 1)
	bottomLeft	= AddVertex (surface, -size, -size, 0, 0, 1)
	AddTriangle (surface, topLeft, bottomRight, bottomLeft)
	AddTriangle (surface, topLeft, topRight, bottomRight)
	Return newMesh
End Function
